from .svg import (
    ColorAndLabel,
    DocumentInfo,
    ParentElement,
    ShapeElement,
    SvgPathCommand,
    curve_to_svg_path,
    edge_to_svg_path,
    edges_from_svg_path,
    face_to_svg_path,
    faces_from_svg_path,
    format_svg,
    import_svg_document,
    wire_to_svg_path,
    wires_from_svg_path,
)
